from pydantic import BaseModel
from humps.camel import case


def to_camel(string):
    return case(string)


class CamelModel(BaseModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True



import os
import pytest
from fa_common.auth import AuthUser

dirname = os.path.dirname(__file__)
env_file = os.path.join(dirname, ".env")
minio_env_file = os.path.join(dirname, ".env-minio")


@pytest.fixture(scope="module")
def env_path(scope="module") -> str:
    clean_env()
    return env_file


@pytest.fixture(scope="module")
def minio_env_path(scope="module") -> str:
    clean_env()
    return minio_env_file


def clean_env():
    os.environ.pop("STORAGE_TYPE", None)
    os.environ.pop("BUCKET_PREFIX", None)
    os.environ.pop("BUCKET_NAME", None)
    os.environ.pop("BUCKET_USER_FOLDER", None)
    os.environ.pop("DATABASE_TYPE", None)
    os.environ.pop("GOOGLE_APPLICATION_CREDENTIALS", None)
    os.environ.pop("ENABLE_FIREBASE", None)
    os.environ.pop("MINIO_ENDPOINT", None)
    os.environ.pop("MINIO_ACCESS_KEY", None)
    os.environ.pop("MINIO_SECRET_KEY", None)


def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        previousfailed = getattr(item.parent, "_previousfailed", None)
        if previousfailed is not None:
            pytest.xfail("previous test failed ({})".format(previousfailed.name))


async def get_test_user() -> AuthUser:
    payload = {
        "sub": "test-sub",
        "name": "Test User",
        "email": "test.user@test.com",
        "nickname": "test.user",
        "email_verified": True,
        "picture": None,
        "updated_at": None,
        "scopes": [],
    }

    return AuthUser(**payload)



from bson import CodecOptions
from motor.motor_asyncio import AsyncIOMotorClient
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.results import DeleteResult, InsertOneResult, UpdateResult

from .models import DocumentDBModel
from fa_common import get_current_app, get_timezone


class MongoDBClient(object):
    """
    Singleton client for interacting with MongoDB.
    Operates mostly using models, specified when making DB queries.

    Implements only part of internal `motor` methods, but can be populated more.

    Please don't use it directly, use `scidra.core.db.utils.get_db_client`.
    """

    __instance = None
    mongodb: AsyncIOMotorClient = None

    def __new__(cls) -> "MongoDBClient":
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            app = get_current_app()
            tzinfo = get_timezone()
            cls.__instance.codec_options = CodecOptions(tz_aware=True, tzinfo=tzinfo)
            cls.__instance.mongodb = app.mongodb
        return cls.__instance

    def get_collection(self, collection_name: str) -> Collection:
        return self.mongodb.get_collection(
            collection_name, codec_options=self.codec_options
        )

    async def insert(
        self,
        model: DocumentDBModel,
        session: ClientSession = None,
        include=None,
        exclude=None,
    ) -> InsertOneResult:
        data = model.dict(include=include, exclude=exclude)
        data["_id"] = data.pop("id")
        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        return await collection.insert_one(data, session=session)

    async def count(
        self, model: DocumentDBModel, session: ClientSession = None, **kwargs
    ) -> int:
        _id = kwargs.pop("id", None)
        if _id is not None:
            kwargs["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.count_documents(kwargs, session=session)
        return res

    async def delete(
        self, model: DocumentDBModel, session: ClientSession = None, **kwargs
    ) -> DeleteResult:
        _id = kwargs.pop("id", None)
        if _id is not None:
            kwargs["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.delete_many(kwargs, session=session)
        return res

    async def update_one(
        self,
        model: DocumentDBModel,
        filter_kwargs: dict,
        session: ClientSession = None,
        **kwargs
    ) -> UpdateResult:
        _id = filter_kwargs.pop("id", None)
        if _id is not None:
            filter_kwargs["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.update_one(filter_kwargs, kwargs, session=session)
        return res

    async def update_many(
        self,
        model: DocumentDBModel,
        filter_kwargs: dict,
        session: ClientSession = None,
        **kwargs
    ) -> UpdateResult:
        _id = filter_kwargs.pop("id", None)
        if _id is not None:
            filter_kwargs["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.update_many(filter_kwargs, kwargs, session=session)
        return res

    async def get(
        self, model: DocumentDBModel, session: ClientSession = None, **kwargs
    ) -> dict:
        _id = kwargs.pop("id", None)
        if _id is not None:
            kwargs["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.find_one(kwargs, session=session)
        return res

    def list(
        self,
        model: DocumentDBModel,
        session: ClientSession = None,
        _offset: int = 0,
        _limit: int = 0,
        _sort: list = None,
        **kwargs
    ) -> Cursor:
        _id = kwargs.pop("id", None)
        if _id is not None:
            kwargs["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        return collection.find(
            kwargs, session=session, skip=_offset, limit=_limit, sort=_sort
        )



from .utils import setup_db, create_indexes, get_db_client
from .models import (
    DBIndex,
    SortOrder,
    FireOffset,
    Operator,
    WhereCondition,
    WriteResult,
    DeleteResult,
    DocumentDBModel,
    DocumentDBTimeStampedModel,
)



from typing import Optional

from fa_common import sizeof_fmt, CamelModel


# Shared properties
class File(CamelModel):
    size: str  # e.g. '3 KB'
    url: Optional[str] = None  # download url
    id: str  # id can be path or database id
    dir: bool  # is current node dir?
    path: str  # path to current item (e.g. /folder1/someFile.txt)
    # optional (but we are using id as name if name is not present) (e.g. someFile.txt)
    name: Optional[str] = None
    content_type: Optional[str]

    def set_size(self, bytes: int):
        self.size = sizeof_fmt(bytes)



from typing import List, Optional

from fastapi import Depends
from fastapi.openapi.models import OAuthFlows
from fastapi.security import OAuth2, SecurityScopes
from jose import JWTError, jwt
from pydantic import ValidationError

from .models import AuthUser
from fa_common import (
    async_get,
    force_async,
    logger as LOG,
    ForbiddenError,
    UnauthorizedError,
    get_settings,
)

oauth2_scheme = OAuth2(
    flows=OAuthFlows(
        implicit={
            "authorizationUrl": get_settings().OAUTH2_AUTH_URL,
            "scopes": {"openid": "", "profile": "", "email": ""},
        }
    )
)


async def get_firebase_token(user: AuthUser) -> str:
    from firebase_admin import auth as fire_auth

    uid = user.sub

    # Potentially add scopes here
    # additional_claims = {
    #     "premiumAccount": True
    # }
    additional_claims: dict = {}
    return await force_async(fire_auth.create_custom_token)(uid, additional_claims)


async def get_user_profile(url: str, auth: str) -> AuthUser:
    data = await async_get(url, auth)
    return AuthUser(**data)


def get_token_auth_header(auth: str) -> str:
    """Obtains the Access Token from the Authorization Header
    """
    if not auth:
        raise UnauthorizedError(detail="Authorization header is expected",)

    parts = auth.split()

    if parts[0].lower() != "bearer":
        raise UnauthorizedError(detail="Authorization header must start with Bearer",)
    elif len(parts) == 1:
        raise UnauthorizedError(detail="Token not found",)
    elif len(parts) > 2:
        raise UnauthorizedError(detail="Authorization header must be Bearer token",)

    token = parts[1]
    return token


async def decode_token(auth: str):
    settings = get_settings()
    token = get_token_auth_header(auth)
    jwks = await async_get(f"https://{settings.AUTH0_DOMAIN}/.well-known/jwks.json")

    unverified_header = jwt.get_unverified_header(token)
    rsa_key: dict = {}
    for key in jwks["keys"]:
        if key["kid"] == unverified_header["kid"]:
            rsa_key = {
                "kty": key["kty"],
                "kid": key["kid"],
                "use": key["use"],
                "n": key["n"],
                "e": key["e"],
            }
    if rsa_key:
        try:
            payload = jwt.decode(
                token,
                rsa_key,
                algorithms=settings.JWT_ALGORITHMS,
                audience=settings.API_AUDIENCE,
                issuer="https://" + settings.AUTH0_DOMAIN + "/",
            )
        except jwt.ExpiredSignatureError:
            raise UnauthorizedError(detail="token is expired",)
        except jwt.JWTClaimsError:
            raise UnauthorizedError(
                detail="incorrect claims, please check the audience and issuer",
            )
        except Exception:
            raise UnauthorizedError(detail="Unable to parse authentication token.",)

        return payload

    raise UnauthorizedError(detail="Unable to find appropriate key",)


async def get_user(payload: dict, token: str) -> Optional[AuthUser]:
    user: Optional[AuthUser] = None
    token_scopes: List[str] = []

    if "permissions" in payload and payload.get("permissions") is not None:
        token_scopes = payload["permissions"]

    if "sub" in payload:
        if (
            "aud" in payload
            and len(payload["aud"]) > 1
            and "userinfo" in payload["aud"][1]
        ):
            user = await get_user_profile(payload["aud"][1], token)
        else:
            user = AuthUser(**payload)

        if len(token_scopes) > 0:
            user.scopes = token_scopes

    return user


async def get_current_user(
    security_scopes: SecurityScopes, token: str = Depends(oauth2_scheme)
) -> AuthUser:

    try:
        payload = await decode_token(token)
        user = await get_user(payload, token)

    except (JWTError, ValidationError) as e:
        LOG.error("Could not validate credentials. {}", str(e))
        raise UnauthorizedError(detail="Could not validate credentials.",)

    if not user:
        raise UnauthorizedError(detail="Invalid authentication credentials.",)

    for scope in security_scopes.scopes:
        if scope not in user.scopes:
            raise ForbiddenError(detail="Not enough permissions",)
    return user
